#ifndef VECTOR_H
#define VECTOR_H
//
//  vector class, copyright 1996, S Zeil, Old Dominion Univ.
//                derived from draft ANSI/ISO C++ standard
//       Permission is granted to copy, distribute and prepare
//       derivative works without charge provided only that
//       this copyright notice is retained.
//
//  This is a "light" version of the standard vector class,
//  designed for use with the variety of compilers currently
//  in use at ODU, including gcc/g++ 2.6.8, Borland Turbo C/C++
//  4.5 (Windows), and Borland Turbo C/C++ 3.1 (MSDOS).
//

#include <algae/config.h>
#include <algae/assert.h>
#include <algae/stdlib.h>

#include <limits.h>

#ifdef PLACEMENT_NEW
#include <new.h>
#define VTYPE T
#define AT(p) *(p)
#ifdef __TCPLUSPLUS__
#include <alloc.h>
#endif
#else
#define VTYPE T*
#define AT(p) *(*(p))
#endif

template <class T>
class vector;

template <class T>
class vector_iterator;


template <class T>
class vector_const_iterator
{
protected:
  const vector<T>* container;
  const VTYPE* position;
  friend class vector<T>;

  vector_const_iterator (const vector<T>* c, const VTYPE* p)
    : container(c), position(p) {}
  friend class vector_iterator<T>;

public:
  vector_const_iterator(): container(0), position(0) {}

  const T& operator*() const  {assert (container != 0); return AT(position);}

  vector_const_iterator& operator++ () // prefix
    {assert ((container != 0)
	     && ((VTYPE*)position < container->data+container->xSize));
     position++;
     return *this;}

  vector_const_iterator operator++ (int)  // postfix: int argument is ignored
    {assert ((container != 0)
	     && ((VTYPE*)position < container->data+container->xSize));
     position++;
     return vector_const_iterator<T>(container, position-1);}

  vector_const_iterator& operator-- () // prefix
    {assert ((container != 0)
	     && ((VTYPE*)position >= container->data));
     position--;
     return *this;}

  vector_const_iterator operator-- (int)  // postfix: int argument is ignored
    {assert ((container != 0)
	     && ((VTYPE*)position >= container->data));
     position--;
     return vector_const_iterator<T>(container, position+1);}

  vector_const_iterator operator+ (unsigned long n) const
    {
      assert (container != 0);
      return vector_const_iterator<T>(container, position+n);
    }
  vector_const_iterator operator- (unsigned long n) const
    {
      assert (container != 0);
      return vector_const_iterator<T>(container, position-n);
    }
  unsigned long operator- (const vector_const_iterator& i) const
    {
      assert ((container != 0) && (container == i.container) 
	      && (position >= i.position));
      return position - i.position;
    }
  bool operator== (const vector_const_iterator& i) const
    {
      assert ((container != 0) && (container == i.container));
      return (position == i.position);
    }
  bool operator!= (const vector_const_iterator& i) const
    {
      assert ((container != 0) && (container == i.container));
      return (position != i.position);
    }
};

template <class T>
class vector_iterator:  public vector_const_iterator<T>
{
private:
  vector_iterator (const vector<T>* c, const VTYPE* p)
    : vector_const_iterator<T>(c,p) {}
  friend class vector<T>;

public:
  vector_iterator(): vector_const_iterator<T>() {}

  T& operator*() const  {return (T&)(AT(position));}

  vector_iterator& operator++ () // prefix
    {assert ((container != 0)
	     && ((VTYPE*)position < container->data+container->xSize));
     position++;
     return *this;}

  vector_iterator operator++ (int)  // postfix: int argument is ignored
    {assert ((container != 0)
	     && ((VTYPE*)position < container->data+container->xSize));
     position++;
     return vector_iterator<T>(container, position-1);}

  vector_iterator& operator-- () // prefix
    {assert ((container != 0)
	     && ((VTYPE*)position >= container->data));
     position--;
     return *this;}

  vector_iterator operator-- (int)  // postfix: int argument is ignored
    {assert ((container != 0)
	     && ((VTYPE*)position >= container->data));
     position--;
     return vector_iterator<T>(container, position+1);}

  vector_iterator operator+ (unsigned long n) const
    {
      assert (container != 0);
      return vector_iterator<T>(container, position+n);
    }
  vector_iterator operator- (unsigned long n) const
    {
      assert (container != 0);
      return vector_iterator<T>(container, position-n);
    }

  unsigned long operator- (const vector_iterator& i) const
    {
      assert ((container != 0) && (container == i.container) 
	      && (position >= i.position));
      return position - i.position;
    }
};




template  <class T>
class vector {
public:
  typedef T* pointer;
  typedef const T* const_pointer;
  typedef T& reference;
  typedef const T& const_reference;
  #define Size_type unsigned int
  typedef Size_type size_type;
  typedef int difference_type;
  #define Iterator vector_iterator<T>
  typedef Iterator iterator;
  #define Const_iterator vector_const_iterator<T>
  typedef Const_iterator const_iterator;

private:
  size_type xSize;
  size_type xCapacity;
  typedef VTYPE ArrayElement;
  VTYPE* data;


  friend class iterator;
  friend class const_iterator;

public:
// Constructors & Destructors

  vector();
  //post: size() == 0

  vector(Size_type n, const T& value =T());
  //  vector v(n, t);
  //post: size() == n &&  ((0 <= i < n) => (v[i] == t))

  vector(const vector& x);
  //  vector v(x);
  //post: v == x

  vector(const_iterator first,
         const_iterator last);
  //  vector v (f, l);
  //post:  (v.size() == l-f) &&
  //       for (p = first; p != last; p++)
  //          *p == v[p-first]

  ~vector();


// Copying

  vector& operator=(const vector& x);
  //  v = x;
  //post:  v == x

  void swap(vector& x);
  //  v = v0;  x = x0;  v.swap(x);
  //post:  v == x0 && x == v0 && ((v0 == x0) == (v == x))

// Iterators

  iterator begin();
  const_iterator begin() const;
  //inv: v.size() > 0  =>  *(v.begin()) == v[0]


  iterator end();
  const_iterator end() const;
  //inv: v.size() > 0  =>  *(v.end()-1) == v[v.size()-1]


// Size and capacity
  size_type size() const;
  size_type max_size() const;
  size_type capacity() const;

  //inv: capacity() <= max_size()
  //inv: size() <= capacity()

  bool empty() const;
  //inv: v.empty() == (v.size() == 0)

  void reserve(Size_type n);
  //pre: n <= max_size()
  //post:  capacity() >= n


// Access

  T& operator[](Size_type n);
  const T& operator[](Size_type n) const;
  //pre: n >= 0 && n < size()

  T& front();
  const T& front() const;
  //pre: size() > 0

  //inv: v.front() == v[0]

  T& back();
  const T& back() const;
  //pre: size() > 0

  //inv: v.back() == v[size()-1]


// Insertion
  void push_back(const T& x);
  //pre: size() < max_size()
  // v0 = v; v.push_back (x);
  //post: v.size() == v0.size()+1
  //      && v.back() == x
  //      && for (i = 0; i < v0.size(); i++)
  //            v0[i] == v[i]

  iterator insert(Iterator position,
                  const T& x);
  //pre: size() < max_size()
  // v0 = v; k = p - v.begin(); v.insert (p, x);
  //post: v.size() == v0.size()+1
  //      && for (i = 0; i < k; i++)
  //            v0[i] == v[i]
  //      && v[k] == x
  //      && for (i = k+1; i < v0.size(); i++)
  //            v0[i] == v[i+1]

  void insert (Iterator position,
               Const_iterator first,
               Const_iterator last);
  //pre: size() + last - first <= max_size()
  // v0 = v; k = p - v.begin(); v.insert (p, f, l);
  //post: v.size() == v0.size()+l-f
  //      && for (i = 0; i < k; i++)
  //            v0[i] == v[i]
  //      && for (i = k; i < k+l-f; i++)
  //            v[i] == *(f+i-k)
  //      && for (i = k+l-f; i < v0.size(); i++)
  //            v0[i] == v[i+l-f]

  void insert (Iterator position,
               Size_type n, const T& x);
  //pre: size() + n <= max_size()
  // v0 = v; k = p - v.begin(); v.insert (p, n x);
  //post: v.size() == v0.size()+n
  //      && for (i = 0; i < k; i++)
  //            v0[i] == v[i]
  //      && for (i = k; i < k+n; i++)
  //            v[i] == x
  //      && for (i = k+n; i < v0.size(); i++)
  //            v0[i] == v[i+n]


// Removal
  void pop_back();
  //pre: size() > 0
  // v0 = v; v.pop_back ();
  //post: v.size() == v0.size()-1
  //      && for (i = 0; i < v.size(); i++)
  //            v0[i] == v[i]

  void erase(Iterator position);
  //pre: size() > 0
  // v0 = v; k = p - v.begin(); v.erase (p);
  //post: v.size() == v0.size()11
  //      && for (i = 0; i < k; i++)
  //            v0[i] == v[i]
  //      && for (i = k+1; i < v.size(); i++)
  //            v0[i] == v[i-1]

  void erase(Iterator first, Iterator last);
  //pre: size() >= last - first
  // v0 = v; k = f - begin(); n = l - f; v.erase (f, l);
  //post: v.size() == v0.size()-n
  //      && for (i = 0; i < k; i++)
  //            v0[i] == v[i]
  //      && for (i = k; i < v.size(); i++)
  //            v0[i+n] == v[i]


private:

  struct T2 {T t1; T t2;};
  struct T3 {T t1; T t2; T t3;};

  void initialize (T*);
};




template <class T>
bool operator== (const vector<T>& left,
		 const vector<T>& right)
{
  if (left.size() == right.size())
    {
      for (int i = 0; i < left.size(); i++)
	if (!(left[i] == right[i]))
	  return false;
      return true;
    }
  else
    return false;
}

template <class T>
bool operator< (const vector<T>& left,
		const vector<T>& right)
{
  vector<T>::size_type sz = left.size();
  if (sz > right.size())
    sz = right.size();

  for (int i = 0; i < sz; i++)
    if (left[i] < right[i])
      return true;
    else if (right[i] < left[i])
      return false;
  return (left.size() < right.size());
}


///////////////////////////////////////////////
#define ValidIterator(p, d) assert(p.container == this \
                                   && (VTYPE*)(p.position) >= (VTYPE*)data \
                                   && (VTYPE*)(p.position) <= (VTYPE*)data + xSize - d)
#define ValidRange(f, l, c) assert(f.container == l.container && \
                                   f.container == c);\
                            assert (c != 0); \
                            assert ((VTYPE*)(f.position) >= (VTYPE*)(c->data));\
                            assert (f.position <= l.position);\
                            assert ((VTYPE*)(l.position) <= (VTYPE*)(c->data)+c->xSize)


// Constructors & Destructors

template <class T>
vector<T>::vector(): xCapacity(0), xSize(0), data(0)
{
  reserve(4);
}


template <class T>
void vector<T>::initialize (T* address)
{
#ifdef PLACEMENT_NEW
  // Force construction of element at address
  new (address) T();
#endif
}

template <class T>
vector<T>::vector(Size_type n, const T& value)
   : xCapacity(0), xSize(0), data(0)
{
  reserve(n);
  xSize = n;
  for (size_type i = 0; i < n; i++)
#ifdef PLACEMENT_NEW
    new (data+i) T(value);
#else
    data[i] = new T(value);
#endif
}


template <class T>
vector<T>::vector(const vector<T>& x): xCapacity(0), xSize(0), data(0)
{
  reserve (x.xSize);
  xSize = x.xSize;
  for (size_type i = 0; i < xSize; i++)
#ifdef PLACEMENT_NEW
    new (data+i) T(x.data[i]);
#else
    data[i] = new T(*(x.data[i]));
#endif
}

template <class T>
vector<T>::vector(Const_iterator first,
               Const_iterator last)
    : xCapacity(0), xSize(0), data(0)
{
  ValidRange (first, last, first.container);
  reserve (last - first);
  xSize = last - first;
  for (size_type i = 0; i < xSize; i++)
#ifdef PLACEMENT_NEW
    new (data+i) T(first.position[i]);
#else
    data[i] = new T(AT(first.position+i));
#endif
}

template <class T>
vector<T>::~vector()
{
#ifdef PLACEMENT_NEW
  for (size_type i = 0; i < xSize; i++)
    (data+i)->~T();
  free (data);
#else
  for (size_type i = 0; i < xSize; i++)
    delete data[i];
  delete [] data;
#endif
}


// Copying

template <class T>
vector<T>& vector<T>::operator=(const vector<T>& x)
{
  if (this != &x)
    {
      if (xSize > x.xSize)
        erase(begin()+x.xSize, end());
      reserve (x.xSize);
      unsigned i;
#ifdef PLACEMENT_NEW
      for (i = xSize; i < x.xSize; i++)
        initialize(data+i);
#endif
      for (i = 0; i < x.xSize; i++)
#ifdef PLACEMENT_NEW
        data[i] = x.data[i];
#else
        if (i >= xSize)
          data[i] = new T(*(x.data[i]));
        else
          *(data[i]) = *(x.data[i]);
#endif
      xSize = x.xSize;
    }
  return *this;
}

template <class T>
void vector<T>::swap(vector<T>& x)
{
  {
    size_type tmp = xCapacity;
    xCapacity = x.xCapacity;
    x.xCapacity = tmp;

    tmp = xSize;
    xSize = x.xSize;
    x.xSize = tmp;
  }
  {
    VTYPE* tmp = data;
    data = x.data;
    x.data = tmp;
  }
}

// Iterators

template <class T>
Iterator vector<T>::begin()
{
  return iterator(this, (const VTYPE*)data);
}

template <class T>
Const_iterator vector<T>::begin() const
{
  return iterator((vector<T>*)this, (const VTYPE*)data);
}


template <class T>
Iterator vector<T>::end()
{
  return iterator(this, (const VTYPE*)data + xSize);
}

template <class T>
Const_iterator vector<T>::end() const
{
  return iterator((vector<T>*)this, (const VTYPE*)data + xSize);
}


// Size and capacity
template <class T>
Size_type vector<T>::size() const
{
  return xSize;
}

template <class T>
Size_type vector<T>::max_size() const
{
  return INT_MAX / (sizeof(T3) - sizeof(T2));
}

template <class T>
Size_type vector<T>::capacity() const
{
  return xCapacity;
}


template <class T>
bool vector<T>::empty() const
{
  return (xSize == 0);
}


#ifdef __TCPLUSPLUS__
template <class T>
void destroy (T* ptr)
{
  ptr->~T();
}
#endif

template <class T>
void vector<T>::reserve(Size_type n)
{
  assert (n <= max_size());
  if (n > xCapacity)
    {
      int k = 4;
      if (n <= max_size()/2)
        while (k < n)  k *= 2;
      else
        k = max_size();
#ifdef PLACEMENT_NEW
      T* newData = (T*)malloc(k * (sizeof(T3) - sizeof(T2)));
      for (int i = 0; i < xSize; i++)
        {
         new (newData+i) T (data[i]);
         (data+i)->~T();
        }
      if (data != 0)
        free (data);
#else
      ArrayElement* newData = new ArrayElement[k];
      for (int i = 0; i < xSize; i++)
        newData[i] = data[i];
      delete [] data;
#endif
      data = newData;
      xCapacity = k;
    }
}


// Access

template <class T>
T& vector<T>::operator[](Size_type n)
{
  assert (n >= 0 && n < xSize);
  return AT(data+n);
}

template <class T>
const T& vector<T>::operator[](Size_type n) const
{
  assert (n >= 0 && n < xSize);
  return AT(data+n);
}

template <class T>
T& vector<T>::front()
{
  assert (xSize > 0);
  return AT(data);
}

template <class T>
const T& vector<T>::front() const
{
  assert (xSize > 0);
  return AT(data);
}


template <class T>
T& vector<T>::back()
{
  assert (xSize > 0);
  return AT(data+xSize-1);
}

template <class T>
const T& vector<T>::back() const
{
  assert (xSize > 0);
  return AT(data+xSize-1);
}



// Insertion
template <class T>
void vector<T>::push_back(const T& x)
{
  assert (xSize < max_size());
  reserve (xSize+1);
#ifdef PLACEMENT_NEW
  new (data+xSize) T(x);
#else
  data[xSize] = new T(x);
#endif
  xSize++;
}

template <class T>
Iterator vector<T>::insert(Iterator position,
                        const T& x)
{
  assert (xSize < max_size());
  ValidIterator (position, 0);
  size_type pos = (VTYPE*)(position.position) - (VTYPE*)data;
  reserve (xSize+1);
#ifdef PLACEMENT_NEW
  initialize(data+xSize);
#endif
  for (size_type p = xSize; p != pos; p--)
    data[p] = data[p-1];
#ifdef PLACEMENT_NEW
  data[pos] = x;
#else
  data[pos] = new T(x);
#endif
  xSize++;
  return position;
}

template <class T>
void vector<T>::insert (Iterator position,
                     Const_iterator first,
                     Const_iterator last)
{
  size_type n = last - first;
  assert ((unsigned long)(xSize + n) < max_size());
  ValidIterator (position, 0);
  ValidRange (first, last, first.container);

  size_type pos = (VTYPE*)(position.position) - (VTYPE*)data;
  reserve (xSize + n);
#ifdef PLACEMENT_NEW
  {
   for (int i = 0; i < n; i++)
     initialize(data+xSize+i);
  }
#endif
  {
   for (size_type p = xSize; p != pos; p--)
     data[p+n-1] = data[p-1];
  }
  {
   VTYPE* p = data + pos;
   for (const_iterator q = first; q != last; q++)
     {
#ifdef PLACEMENT_NEW
       *p = *q;
#else
       *p = new T(*q);
#endif
       p++;
     }
  }
  xSize += n;
}

template <class T>
void vector<T>::insert (Iterator position,
                     Size_type n, const T& x)
{
  assert ((unsigned long)(xSize + n) < max_size());
  ValidIterator (position, 0);
  size_type pos = (VTYPE*)(position.position) - (VTYPE*)data;
  reserve (xSize + n);
#ifdef PLACEMENT_NEW
  {
   for (int i = 0; i < n; i++)
     initialize(data+xSize+i);
  }
#endif
  {
   for (size_type p = xSize; p != pos; p--)
     data[p+n-1] = data[p-1];
  }
  {
   VTYPE* p = data + pos;
   for (size_type q = 0; q < n; q++)
     {
#ifdef PLACEMENT_NEW
       *p = x;
#else
       *p = new T(x);
#endif
       p++;
     }
  }
  xSize += n;
}


// Removal
template <class T>
void vector<T>::pop_back()
{
  assert (xSize > 0);
  xSize--;
#ifdef PLACEMENT_NEW
  (data+xSize)->~T();
#else
  delete data[xSize];
#endif
}

template <class T>
void vector<T>::erase(Iterator position)
{
  assert (xSize > 0);
  ValidIterator (position, 1);

  for (iterator p = position+1; p != end(); p++)
    *(p-1) = *p;
  xSize--;
#ifdef PLACEMENT_NEW
  (data+xSize)->~T();
#else
  delete data[xSize];
#endif
}

template <class T>
void vector<T>::erase(Iterator first, Iterator last)
{
  assert (xSize >= last - first);
  ValidRange(first, last, this);
  size_type ifirst = (VTYPE*)(first.position) - (VTYPE*)data;
  size_type ilast = (VTYPE*)(last.position) - (VTYPE*)data;
#ifndef PLACEMENT_NEW
  {
   for (int i = ifirst; i < ilast; ++i)
     delete data[i];
  }
#endif
  int p = ifirst;
  for (int q = ilast; q < xSize; q++)
    {
      data[p] = data[q];
      p++;
    }
  xSize -= (last - first);
#ifdef PLACEMENT_NEW
  for (T* t = data+xSize; t != data+xSize+(last-first); t++)
    t->~T();
#endif
}


#undef Size_type
#undef Iterator
#undef Const_iterator

#endif

