/**
 * Client.java
 *
 *
 * Created: Sat Apr 25 22:14:01 1998
 *
 * @author Steven J. Zeil
 * @version
 */

package edu.odu.cs.zeil.AlgAE.Client;

import java.awt.BorderLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowEvent;
import java.net.URL;
import java.util.LinkedList;
import java.util.List;

import javax.help.CSH;
import javax.help.HelpBroker;
import javax.help.HelpSet;
import javax.swing.JFrame;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;

import edu.odu.cs.zeil.AlgAE.Animation;
import edu.odu.cs.zeil.AlgAE.AnimationContext;
import edu.odu.cs.zeil.AlgAE.ContextAware;
import edu.odu.cs.zeil.AlgAE.Client.DataViewer.Animator;
import edu.odu.cs.zeil.AlgAE.Client.IOViewer.IOPane;
import edu.odu.cs.zeil.AlgAE.Client.SourceViewer.SourcePane;
import edu.odu.cs.zeil.AlgAE.Snapshot.Snapshot;


public class Client extends Anchors implements ContextAware {

	private ViewerPanel dataViewer;
	private Animator animator;
	private List<JMenuItem> algorithmItems;
	
	private JFrame mainWindow;
	private JMenuBar menuBar;
	private JMenu algorithmMenu;
	private Animation animation;
	private ClientThread messageReader;
	private IOPane ioPane;

	private HelpSet hs;
	private HelpBroker hb;

	/**
	 * Indicates whether we are running some portion of the
	 * animated code.
	 */
	private boolean running;

	
	/**
	 * Indicates that animation has been shut down
	 */
	private boolean stopped;

	/**
	 * Indicates whether animation is in its registered starting action
	 */
	private boolean starting;

	/**
	 * Indicates whether animation is awaiting its first algorithm selection
	 */
	private boolean firstSelection;

	  
	public static final String StartingAction = "*start*";
	
	  
	public Client(Animation anim, JMenuBar menuBar, JFrame mainWindow, boolean startPausing)
	{
		super();
		
		this.mainWindow = mainWindow;
		this.animation = anim;
		this.menuBar = menuBar;
		
		running = false;
		stopped = false;
		starting = false;
		firstSelection = false;
		algorithmItems = new LinkedList<JMenuItem>();
		
		setLayout (new BorderLayout());
				
				
		ioPane = new IOPane (anim);
		SourcePane sourcePane = new SourcePane();
		
		animator = new Animator(sourcePane, this);
		dataViewer = new ViewerPanel ("AlgAE", animator, sourcePane, ioPane);
		
		add (dataViewer, BorderLayout.CENTER);
		
		messageReader = null;
	}







	private void buildMenu(JMenuBar menuBar) {

		if (mainWindow != null) { 
			JMenu fileMenu = new JMenu ("File", false);
			menuBar.add(fileMenu);

			JMenuItem exitItem = new JMenuItem ("Exit");
			fileMenu.add (exitItem);
			exitItem.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					mainWindow.dispatchEvent(new WindowEvent(mainWindow, WindowEvent.WINDOW_CLOSING));
					mainWindow.setVisible(false);
				}});
		}

		algorithmMenu = new JMenu ("Algorithm", false);
		menuBar.add (algorithmMenu);


		JMenu helpMenu = new JMenu ("Help", false);
		menuBar.add(helpMenu);



		animation.buildMenu();


		JMenuItem helpItem = new JMenuItem ("Help");
		helpMenu.add (helpItem);

		String helpHS = "Help/helpset.hs";
		ClassLoader cl = Client.class.getClassLoader();
		try {
			URL hsURL = HelpSet.findHelpSet(cl, helpHS);
			hs = new HelpSet(null, hsURL);
		} catch (Exception ee) {
			System.err.println( "HelpSet " + ee.getMessage());
		}
		if (hs != null) {
			hb = hs.createHelpBroker();

			helpItem.addActionListener(new CSH.DisplayHelpFromSource (hb));
		}
		
		
		JMenuItem aboutItem = new JMenuItem ("About AlgAE");
		helpMenu.add (aboutItem);
		aboutItem.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				JOptionPane.showMessageDialog(mainWindow,
						animation.about() +
						"\n----------------------------\n"
						+ "AlgAE Algorithm Animation Engine, " +
						"version 3.0\n\n" +
						"  Steven J. Zeil\n" +
						"  Old Dominion University\n" +
				"  Dept. of Computer Science");
			}});
		
	}





	public void start() {
		buildMenu (menuBar);
		menuBar.invalidate();
		messageReader = new ClientThread(this);
		messageReader.setPriority(Thread.MIN_PRIORITY);
		messageReader.start();
		menuBar.invalidate();
	}

	/**
	 * Called from Animation applet to indicate that a start() call
	 * has been issued.
	 */
	public void resume() {
		animator.resumeAnimator();
	}

	/**
	 * Called from Animation applet to indicate that a stop() call
	 * has been issued.
	 */
	public void pause() {
		animator.pauseAnimator();
	}

	/**
	 * Called from Animation applet to indicate that a destroy() call
	 * has been issued.
	 */
	public void destroy() {
		if (!stopped) {
			animator.shutdown();
			animation.getServer().shutdown();
			stopped = true;
			try {
				Thread.sleep(500);
				if (messageReader.isAlive()) {
					messageReader.interrupt();
				}
			} catch (Exception e1) {
			}
		}
	}


	private ActionListener algorithmMenuAction = new ActionListener() {
		
		@Override
		public void actionPerformed(ActionEvent ev) {
			JMenuItem selected = (JMenuItem)ev.getSource();
			String name = selected.getText();
			algorithmMenuItemSelected(name);
		}
	};
	
	public void registerMenuItem(String menuItemTitle) {
		JMenuItem newItem = new JMenuItem (menuItemTitle);
		algorithmMenu.add(newItem);
		algorithmItems.add(newItem);
		newItem.addActionListener(algorithmMenuAction);
	}


	protected void enableAlgorithmMenuItems(boolean b) {
		for (JMenuItem item: algorithmItems) {
			item.setEnabled(b);
		}
		algorithmMenu.setEnabled(b);
	}


	public void showSnapshot(Snapshot snapshot) throws InterruptedException {
		animator.add(snapshot);
	}

	
	private class ClientThread extends Thread {

		private Client client;

		public ClientThread(Client client) {
			super("ClientThread");
			this.client = client;
		}


		public void run() {
			synchronized (client) {
				starting = true;
				enableAlgorithmMenuItems (false);
				algorithmMenuItemSelected(StartingAction);
				animator.startPlay();
			}

			while (!stopped) {
				Thread.yield();
				Message msg;
				try {
					//System.err.println("Client thread " + paused + " " + messages.size());
					msg = animation.getMessageQueue().take();
					//System.err.println("Client thread " + paused);
				} catch (InterruptedException e) {
					break;
				}
				try {
					msg.interpret (client);
				} catch (InterruptedException e) {
					break;
				}
			}
		}
	}

	protected void algorithmMenuItemSelected(String name) {
		if (!running) {
			enableAlgorithmMenuItems (false);
			running = true;
			animator.clear();
			if (starting) {
				starting = false;
				firstSelection = true;
				animator.startPlay();
			} else if (firstSelection) {
				animator.startStepping();
				firstSelection = false;
		}	
			animation.getServer().menuItemSelected (name);
		}
	}


	public void menuItemCompleted(Snapshot snapshot) {
		try {
			animator.add(snapshot);
		} catch (InterruptedException e) {
		}
		running = false;
		animator.showStatus("Choose an Algorithm");
		animator.endofSequence();
		enableAlgorithmMenuItems (true);
	}






	public void promptForInputDialog(String prompt, String requiredPattern) {
		String response = JOptionPane.showInputDialog(this, prompt, "Input Requested", JOptionPane.QUESTION_MESSAGE);
		while (!response.matches(requiredPattern)) {
			response = JOptionPane.showInputDialog(this, prompt, "Try again", JOptionPane.WARNING_MESSAGE);
		}
		animation.getServer().inputSupplied(response);
	}


	public AnimationContext context() {
		return animation;
	}







	public IOPane getIOPane() {
		return ioPane;
	}







	/**
	 * Provide access to the animator
	 */
	public Animator getAnimator() {
		return animator;
	}







}
