package edu.odu.cs.zeil.AlgAE.Server;

import java.lang.ref.WeakReference;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.concurrent.ArrayBlockingQueue;

import edu.odu.cs.zeil.AlgAE.Animation;
import edu.odu.cs.zeil.AlgAE.AnimationContext;
import edu.odu.cs.zeil.AlgAE.ContextAware;
import edu.odu.cs.zeil.AlgAE.Client.MenuItemCompletedMessage;
import edu.odu.cs.zeil.AlgAE.Client.Message;
import edu.odu.cs.zeil.AlgAE.Client.PromptForInputMessage;
import edu.odu.cs.zeil.AlgAE.Client.IOViewer.SimulatedPrintStream;
import edu.odu.cs.zeil.AlgAE.Client.SourceViewer.SourceLocation;
import edu.odu.cs.zeil.AlgAE.Snapshot.ActivationStack;
import edu.odu.cs.zeil.AlgAE.Snapshot.Identifier;
import edu.odu.cs.zeil.AlgAE.Snapshot.Snapshot;


public class Server implements ContextAware
{
  /**
   *  Messages to be sent to the client.
   */
  private ArrayBlockingQueue<Message> messages;
  
  /**
   * Animation instance
   */
  private Animation animation;
  
  /**
   *  Collection of algorithm menu items.
   **/
  private HashMap<String, MenuFunction> algorithmsMenu;

  
  /**
   * The algorithms being animated are run from a separate thread.
   */
  private ServerThread thread;
  
  
  /**
   * Most recent selection from the algorithms menu
   */
  private String menuSelection;
  
  
  /**
   * Used to force stop of thread
   */
  private boolean stopped;

  /**
   * All known objects in the code being animated, indexed by unique Identifiers
   * assigned by this class.
   */
  private HashMap<Identifier, WeakReference<Object> > knownObjects;

private String promptedInput;

  /**
   *  Constructor for AlgAE server.
   *  @param anim  Animation isntance for which this is a server
   */
  public Server(Animation anim)
  {
	  animation = anim;
	  knownObjects= new HashMap<Identifier, WeakReference<Object> >();
	  messages = anim.getMessageQueue();
	  SimulatedPrintStream.setMsgQueue(messages);
	  algorithmsMenu = new HashMap<String, MenuFunction>();
	  menuSelection = null;
	  thread = new ServerThread();
	  anim.registerInstance(thread);
	  stopped = false;
  }

  /**
   *  Declare a menu item to appear in the client's Algorithm menu
   *  and a corresponding function to be executed here on the server
   *  in response to a selection of that item.
   *
   * @param menuItemTitle The string to appear in the client's menu
   * @param theOperation  code to invoke when the menu item is selected
   **/
  public void registerMenuItem(String menuItemTitle, MenuFunction theOperation)
  {
    algorithmsMenu.put (menuItemTitle, theOperation);
  }

  /**
   * Start the client and begin running the animation server.
   *  @param animationTitle   string to place in client's window title bar
   **/
  public void start()
  {
//    msgHandler.algorithmName (animationTitle, menuGenerator.about());
//    
//
//    // Main loop: process client's menu selections
//    String menuSelection = "";
//    while (!menuSelection.equals("QUIT"))
//      {
//	msgHandler.beginAlgorithm();
//	describeData();
//	msgHandler.status ("Select an algorithm");
//	msgHandler.endAlgorithm();
//	menuSelection = msgHandler.getMenuSelection ();
//	invoke (menuSelection);
//      }
	  thread.setPriority(Thread.MIN_PRIORITY);
	  thread.start();
  }


/**
 * Called from client to indicate that a selection was made from
 * the algorithms menu.
 * 
 * @param name
 */
public void menuItemSelected(String name) {
	synchronized (algorithmsMenu) {
		menuSelection = name;
		algorithmsMenu.notifyAll();
	}
	//System.err.println ("notified server thread");	
}

/**
 * Pops up a dialog box prompting for an input, pausing the
 * animation until a satisfactory input value is obtained from the
 * human operator.
 * 
 * @param prompt  Text of the prompt message to be displayed
 * @param requiredPattern regular expression describing an acceptable input value
 * @return a human-entered string matching the requiredPattern
 */
public String promptForInput(String prompt, String requiredPattern) {
	promptedInput = "";
	Message promptMsg = new PromptForInputMessage(prompt, requiredPattern);
	try {
		messages.put(promptMsg);
		synchronized (messages) {
			messages.wait();
			return promptedInput;
		}
	} catch (InterruptedException e) {
		return "";
	}

}

/**
 * Provide an input from a dialog box popped up in a separate thread.ops up a dialog box prompting for an input, pausing the
 * animation until a satisfactory input value is obtained from the
 * human operator.
 * 
 * @param value  input text
 */
public void inputSupplied(String value) {
	synchronized (messages) {
		promptedInput = value;
		messages.notifyAll();
	}
}




/**
 * Try to find a known identifier for this object. If none is found, register
 * a new, arbitrary identifier for the purpose.
 *  
 * @param c
 * @return the unique ID for this object
 */
public  Identifier getIdentifierFor(Object obj) {
	if (obj == null)
		return Identifier.nullID;
	LinkedList<Identifier> garbage = new LinkedList<Identifier>();
	Identifier result = null;
	for (Identifier knownOid: knownObjects.keySet()) {
		WeakReference<Object> knownObjW = knownObjects.get(knownOid);
		Object knownObj = knownObjW.get();
		if (knownObj == obj) {
			result = knownOid;
			break;
		} else if (knownObj == null) {
			garbage.add (knownOid);
		}
	}
	for (Identifier garbageId: garbage) {
		knownObjects.remove(garbageId);
	}
	if (result == null) {
		result = new Identifier();
		knownObjects.put (result, new WeakReference<Object>(obj));
	}
	return result;
}


  
private class ServerThread extends Thread {
	
	public ServerThread() {
		super("Server Thread");
	}
	
	public void run() {
		//System.err.println ("Started server thread");
		String selection = null;
		while (!stopped) {
			Thread.yield();
			synchronized (algorithmsMenu) {
				while (menuSelection == null) {
					try {
						algorithmsMenu.wait();
					} catch (InterruptedException e) {
						break;
					}
				}
				selection = menuSelection;
				menuSelection = null;
			}
			MenuFunction mf = algorithmsMenu.get(selection);
			if (mf != null) {
				mf.selected();
			}
			ActivationStack astack = animation.getActivationStack();
			astack.clear();
			Snapshot snap = new Snapshot("Choose an Algorithm", new SourceLocation(getClass(), "", 0),
					astack);
			Message menuSelectionCompleted = new MenuItemCompletedMessage(snap);
			try {
				messages.put(menuSelectionCompleted);
			} catch (InterruptedException e) {
				break;
			}
		}
			
	}
}



public void shutdown() {
	stopped = true;
	try {
		Thread.sleep(500);
		if (thread.isAlive()) {
			thread.interrupt();
		}
	} catch (Exception e) {
		
	}
	
}

@Override
public AnimationContext context() {
	return animation;
}
  
  
  

  
}
